@@ -24,6 +24,7 @@ from __future__ import absolute_import, division, print_function, unicode_litera
 
 from logging import DEBUG as _DEBUG
 from logging import INFO as _INFO
+from logging import WARNING as _WARNING
 from logging import getLogger
 
 from . import hidpp10 as _hidpp10
@@ -102,6 +103,12 @@ def _process_device_notification(device, status, n):
     # HID++ 1.0 requests, should never get here
     assert n.sub_id & 0x80 == 0
 
+    # Allow the device object to handle the notification using custom
+    # per-device state.
+    handling_ret = device.handle_notification(n)
+    if handling_ret is not None:
+        return handling_ret
+
     # 0x40 to 0x7F appear to be HID++ 1.0 or DJ notifications
     if n.sub_id >= 0x40:
         if len(n.data) == _DJ_NOTIFICATION_LENGTH:
@@ -130,7 +137,7 @@ def _process_device_notification(device, status, n):
 
 def _process_dj_notification(device, status, n):
     if _log.isEnabledFor(_DEBUG):
-        _log.debug('%s (%s) DJ notification %s', device, device.protocol, n)
+        _log.debug('%s (%s) DJ %s', device, device.protocol, n)
 
     if n.sub_id == 0x40:
         # do all DJ paired notifications also show up as HID++ 1.0 notifications?
@@ -145,8 +152,10 @@ def _process_dj_notification(device, status, n):
         return True
 
     if n.sub_id == 0x42:
+        connected = not n.address & 0x01
         if _log.isEnabledFor(_INFO):
-            _log.info('%s: ignoring DJ connection: %s', device, n)
+            _log.info('%s: DJ connection: %s %s', device, connected, n)
+        status.changed(active=connected, alert=_ALERT.NONE, reason=_('connected') if connected else _('disconnected'))
         return True
 
     _log.warn('%s: unrecognized DJ %s', device, n)
@@ -199,13 +208,12 @@ def _process_hidpp10_notification(device, status, n):
             0x0A else 'Lightspeed 1' if n.address == 0x0C else 'Lightspeed 1_1' if n.address == 0x0D else None
         )
         if protocol_name:
-            if _log.isEnabledFor(_DEBUG):
-                wpid = _strhex(n.data[2:3] + n.data[1:2])
-                # workaround for short EX100 wpids
-                if protocol_name == '27 MHz':
-                    wpid = _strhex(n.data[2:3]) + '00'
-                assert wpid == device.wpid, '%s wpid mismatch, got %s' % (device, wpid)
-
+            wpid = _strhex(n.data[2:3] + n.data[1:2])
+            # workaround for short EX100 and other 27 MHz wpids
+            if protocol_name == '27 MHz':
+                wpid = '00' + _strhex(n.data[2:3])
+            if wpid != device.wpid:
+                _log.warn('%s wpid mismatch, got %s', device, wpid)
             flags = ord(n.data[:1]) & 0xF0
             link_encrypted = bool(flags & 0x20)
             link_established = not (flags & 0x40)
@@ -252,14 +260,17 @@ def _process_feature_notification(device, status, n, feature):
             discharge_next_level = ord(n.data[1:2])
             battery_status = ord(n.data[2:3])
             status.set_battery_info(discharge_level, _hidpp20.BATTERY_STATUS[battery_status], discharge_next_level)
+        elif n.address == 0x10:
+            if _log.isEnabledFor(_INFO):
+                _log.info('%s: spurious BATTERY status %s', device, n)
         else:
             _log.warn('%s: unknown BATTERY %s', device, n)
         return True
 
     if feature == _F.BATTERY_VOLTAGE:
         if n.address == 0x00:
-            level, status, voltage, _ignore, _ignore = _hidpp20.decipher_voltage(n.data)
-            status.set_battery_info(level, status, None, voltage)
+            battery_level, battery_status, battery_voltage, _ignore, _ignore = _hidpp20.decipher_voltage(n.data)
+            status.set_battery_info(battery_level, battery_status, None, battery_voltage)
         else:
             _log.warn('%s: unknown VOLTAGE %s', device, n)
         return True
@@ -270,15 +281,38 @@ def _process_feature_notification(device, status, n, feature):
             if _log.isEnabledFor(_INFO):
                 _log.info('%s: reprogrammable key: %s', device, n)
         else:
-            _log.warn('%s: unknown REPROGRAMMABLE KEYS %s', device, n)
+            _log.warn('%s: unknown REPROG_CONTROLS %s', device, n)
         return True
 
+    if feature == _F.REPROG_CONTROLS_V4:
+        if n.address == 0x00:
+            if _log.isEnabledFor(_DEBUG):
+                cid1, cid2, cid3, cid4 = _unpack('!HHHH', n.data[:8])
+                _log.debug('%s: diverted controls pressed: %i, %i, %i, %i', device, cid1, cid2, cid3, cid4)
+            return True
+        elif n.address == 0x10:
+            if _log.isEnabledFor(_DEBUG):
+                dx, dy = _unpack('!hh', n.data[:4])
+                _log.debug('%s: rawXY dx=%i dy=%i', device, dx, dy)
+            return True
+        elif n.address == 0x20:
+            if _log.isEnabledFor(_DEBUG):
+                _log.debug('%s: received analyticsKeyEvents', device)
+            return True
+        elif _log.isEnabledFor(_WARNING):
+            _log.warn('%s: unknown REPROG_CONTROLS_V4 %s', device, n)
+
     if feature == _F.WIRELESS_DEVICE_STATUS:
         if n.address == 0x00:
             if _log.isEnabledFor(_DEBUG):
                 _log.debug('wireless status: %s', n)
-            if n.data[0:3] == b'\x01\x01\x01':
-                status.changed(active=True, alert=_ALERT.NOTIFICATION, reason='powered on')
+            reason = 'powered on' if n.data[2] == 1 else None
+            if n.data[1] == 1:  # device is asking for software reconfiguration so need to change status
+                # only show a user notification if the device can change hosts
+                # as we want to notify when a device changes to this host
+                # but the only indication we get is this notification
+                alert = _ALERT.NOTIFICATION if _F.CHANGE_HOST in device.features else _ALERT.NONE
+                status.changed(active=True, alert=alert, reason=reason)
             else:
                 _log.warn('%s: unknown WIRELESS %s', device, n)
         else:
